/**********************************************************
**   PARSER_C.FRM
**   Coco/R C Support Frames.
**   Author: Frankie Arzu <farzu@uvg.edu.gt>
**
**   Jun 12, 1996   Version 1.06
**      Many fixes and suggestions thanks to
**      Pat Terry <p.terry@ru.ac.za>
**********************************************************/

#include "tastes.h"
#include "tastep.h"

#include <string.h>
#include "tl.h"
#include "tc.h"

typedef char Name[15];

void StringToVal (char *s, int *val)
{ int n = 0;
  while (*s) n = 10 * n + (*s++ - '0');
  *val = n;
}

/*--------------------------------------------------------------------------*/



Error_Func Custom_Error = 0L;

static int Sym;
static int errors = 0;                /*number of detected errors*/
static int ErrDist = MinErrDist;

#define MAXSYM		2


/* Production prototypes */

static void Taste(void);
static void Ident(Name name);
static void Body(void);
static void TypeId(int *type);
static void StatSeq(void);
static void Stat(void);
static void Expression(int *type);
static void SimExpr(int *type);
static void RelOp(int *op);
static void Term(int *type);
static void AddOp(int *op);
static void Factor(int *type);
static void MulOp(int *op);


#define NSETBITS        16

static unsigned short int SymSet[][MAXSYM] = {
  /*EOF_Sym */
  {0x1,0x0},
  {0x0}
};

void GenError(int errno)
{ if (ErrDist >= MinErrDist) {
    if (Custom_Error != 0L)
      (*Custom_Error) (errno, S_NextLine, S_NextCol, S_NextPos);
    errors++;
  }
  ErrDist = 0;
}

void SynError(int errno)
{ if (errno <= MAXERROR) errno = MAXERROR;
  if (ErrDist >= MinErrDist) {
    if (Custom_Error != 0L)
      (*Custom_Error) (errno, S_NextLine, S_NextCol, S_NextPos);
    errors++;
  }
  ErrDist = 0;
}

void SemError(int errno)
{ if (errno <= MAXERROR) errno = MAXERROR;
  if (ErrDist >= MinErrDist) {
    if (Custom_Error != 0L)
      (*Custom_Error) (errno, S_Line, S_Col, S_Pos);
    errors++;
  }
  ErrDist = 0;
}

static void Get(void)
{ do {
    Sym = S_Get();
    if (Sym <= MAXT) ErrDist ++;
    else {
      /* Empty Stmt */ ;
      S_NextPos  = S_Pos;  S_NextCol = S_Col;
      S_NextLine = S_Line; S_NextLen = S_Len;
    }
  } while (Sym > MAXT);
}

static int In (unsigned short int *SymbolSet, int i)
{ return SymbolSet[i / NSETBITS] & (1 << (i % NSETBITS)); }

static void Expect (int n)
{ if (Sym == n) Get(); else GenError(n); }

static void ExpectWeak (int n, int follow)
{ if (Sym == n) Get();
  else {
    GenError(n);
    while (!(In(SymSet[follow], Sym) || In(SymSet[0], Sym))) Get();
  }
}

static int WeakSeparator (int n, int syFol, int repFol)
{ unsigned short int s[MAXSYM];
  int i;

  if (Sym == n) { Get(); return 1; }
  if (In(SymSet[repFol], Sym)) return 0;
  for (i = 0; i < MAXSYM; i++)
    s[i] = SymSet[0][i] | SymSet[syFol][i] | SymSet[repFol][i];
  GenError(n);
  while (!In(s, Sym)) Get();
  return In(SymSet[syFol], Sym);
}

int Successful(void)
{ return errors == 0; }

/* Productions */

static void Taste(void)
{
	Name name, progName;
	Object obj;
	tl_init();
	Expect(PROGRAMSym);
	Ident(progName);
	Expect(SemicolonSym);
	progStart = pc;
	Body();
	Ident(name);
	if (strcmp(name, progName) != 0)
	  SemError(119);
	Emit(HALTc);
	Expect(PointSym);
}

static void Ident(Name name)
{
	Expect(identSym);
	LexName(name, sizeof(Name)-1);
}

static void Body(void)
{
	int fix, type;
	Name name, name1;
	Object obj;
	EnterScope();
	fix = pc + 1; Emit2(JMP, 0);
	while (Sym == VARSym ||
	       Sym == PROCEDURESym) {
		if (Sym == VARSym) {
			Get();
			while (Sym == identSym) {
				Ident(name);
				Expect(ColonSym);
				obj = NewObj(name, VARS);
				TypeId(&(*obj).type);
				Expect(SemicolonSym);
			}
		} else if (Sym == PROCEDURESym) {
			Get();
			Ident(name);
			Expect(SemicolonSym);
			obj = NewObj(name, PROCS);
			obj->adr = pc;
			Body();
			Ident(name1);
			Emit(RET);
			if (strcmp(name, name1) != 0)
			  SemError(119);
			Expect(SemicolonSym);
		} else GenError(30);
	}
	Expect(BEGINSym);
	Fixup(fix); Emit2(RES, DataSpace());
	StatSeq();
	Expect(ENDSym);
	LeaveScope();
}

static void TypeId(int *type)
{
	*type = UNDEF;
	if (Sym == INTEGERSym) {
		Get();
		*type = INT;
	} else if (Sym == BOOLEANSym) {
		Get();
		*type = BOOL;
	} else GenError(31);
}

static void StatSeq(void)
{
	Stat();
	while (Sym == SemicolonSym) {
		Get();
		Stat();
	}
}

static void Stat(void)
{
	int  type;
	Name name;
	Object obj;
	int fix, fix2, loopstart;
	if (Sym == identSym ||
	    Sym == IFSym ||
	    Sym == WHILESym ||
	    Sym >= READSym && Sym <= WRITESym) {
		switch (Sym) {
			case identSym:  
				Ident(name);
				obj = Obj(name);
				if (Sym == ColonSym) {
					Get();
					Expect(EqualSym);
					if (obj->kind != VARS) SemError(123);
					Expression(&type);
					if (type != obj->type) SemError(121);
					Emit3(STO, curLevel-obj->level, obj->adr);
				} else if (Sym == SemicolonSym ||
				           Sym == ENDSym ||
				           Sym == ELSESym) {
					if (obj->kind != PROCS) SemError(124);
					Emit3(CALL, curLevel-obj->level, obj->adr);
				} else GenError(32);
				break;
			case IFSym:  
				Get();
				Expression(&type);
				if (type != BOOL) SemError(122);
				fix = pc + 1; Emit2(FJMP, 0);
				Expect(THENSym);
				StatSeq();
				if (Sym == ELSESym) {
					Get();
					fix2 = pc + 1; Emit2(JMP, 0);
					Fixup(fix); fix = fix2;
					StatSeq();
				}
				Expect(ENDSym);
				Fixup(fix);
				break;
			case WHILESym:  
				Get();
				loopstart = pc;
				Expression(&type);
				if (type != BOOL) SemError(122);
				fix = pc + 1; Emit2(FJMP, 0);
				Expect(DOSym);
				StatSeq();
				Emit2(JMP, loopstart); Fixup(fix);
				Expect(ENDSym);
				break;
			case READSym:  
				Get();
				Ident(name);
				obj = Obj(name);
				if (obj->type != INT) SemError(120);
				Emit3(READ, curLevel-obj->level, obj->adr);
				break;
			case WRITESym:  
				Get();
				Expression(&type);
				if (type != INT) SemError(120);
				Emit(WRITE);
				break;
			default :GenError(33); break;
		}
	}
}

static void Expression(int *type)
{
	int type1, op;
	SimExpr(type);
	if (Sym == EqualSym ||
	    Sym >= LessSym && Sym <= GreaterSym) {
		RelOp(&op);
		SimExpr(&type1);
		if (*type != type1) SemError(121);
		Emit(op); *type = BOOL;
	}
}

static void SimExpr(int *type)
{
	int type1, op;
	Term(type);
	while (Sym == MinusSym ||
	       Sym == PlusSym) {
		AddOp(&op);
		Term(&type1);
		if (*type != INT || type1 != INT) SemError(120);
		Emit(op);
	}
}

static void RelOp(int *op)
{
	*op = -1;
	if (Sym == EqualSym) {
		Get();
		*op = EQU;
	} else if (Sym == LessSym) {
		Get();
		*op = LSS;
	} else if (Sym == GreaterSym) {
		Get();
		*op = GTR;
	} else GenError(34);
}

static void Term(int *type)
{
	int type1, op;
	Factor(type);
	while (Sym >= StarSym && Sym <= SlashSym) {
		MulOp(&op);
		Factor(&type1);
		if (*type != INT || type1 != INT) SemError(120);
		Emit(op);
	}
}

static void AddOp(int *op)
{
	*op = -1;
	if (Sym == PlusSym) {
		Get();
		*op = PLUS;
	} else if (Sym == MinusSym) {
		Get();
		*op = MINUS;
	} else GenError(35);
}

static void Factor(int *type)
{
	int val, n;
	Object obj;
	Name name;
	switch (Sym) {
		case identSym:  
			Ident(name);
			obj = Obj(name); *type = obj->type;
			if (obj->kind == VARS)
			  Emit3(LOAD, curLevel-obj->level, obj->adr);
			else SemError(123);
			break;
		case TRUESym:  
			Get();
			Emit2(LIT, 1); *type = BOOL;
			break;
		case FALSESym:  
			Get();
			Emit2(LIT, 0); *type = BOOL;
			break;
		case numberSym:  
			Get();
			LexName(name, sizeof(name)-1);
			StringToVal(name, &n);
			Emit2(LIT, n); *type = INT;
			break;
		case MinusSym:  
			Get();
			Factor(type);
			if (*type != INT) { SemError(120); *type = INT; }
			Emit(NEG);
			break;
		default :GenError(36); break;
	}
}

static void MulOp(int *op)
{
	*op = -1;
	if (Sym == StarSym) {
		Get();
		*op = TIMES;
	} else if (Sym == SlashSym) {
		Get();
		*op = SLASH;
	} else GenError(37);
}



void Parse(void)
{ S_Reset(); Get();
  Taste();
}

