/**********************************************************
**   PARSER_C.FRM
**   Coco/R C Support Frames.
**   Author: Frankie Arzu <farzu@uvg.edu.gt>
**
**   Jun 12, 1996   Version 1.06
**      Many fixes and suggestions thanks to
**      Pat Terry <p.terry@ru.ac.za>
**********************************************************/

#include "xrefs.h"
#include "xrefp.h"

#include "crossref.h"
#define  APPLIED  0
#define  DEFINED  1

static  char Name[15];

/*------------------------------------------------------------------------*/



Error_Func Custom_Error = 0L;

static int Sym;
static int errors = 0;                /*number of detected errors*/
static int ErrDist = MinErrDist;

#define MAXSYM		2


/* Production prototypes */

static void XRef(void);
static void Ident(void);
static void Body(void);
static void TypeId(void);
static void StatSeq(void);
static void Stat(void);
static void Expression(void);
static void SimExpr(void);
static void RelOp(void);
static void Term(void);
static void AddOp(void);
static void Factor(void);
static void MulOp(void);


#define NSETBITS        16

static unsigned short int SymSet[][MAXSYM] = {
  /*EOF_Sym */
  {0x1,0x0},
  {0x0}
};

void GenError(int errno)
{ if (ErrDist >= MinErrDist) {
    if (Custom_Error != 0L)
      (*Custom_Error) (errno, S_NextLine, S_NextCol, S_NextPos);
    errors++;
  }
  ErrDist = 0;
}

void SynError(int errno)
{ if (errno <= MAXERROR) errno = MAXERROR;
  if (ErrDist >= MinErrDist) {
    if (Custom_Error != 0L)
      (*Custom_Error) (errno, S_NextLine, S_NextCol, S_NextPos);
    errors++;
  }
  ErrDist = 0;
}

void SemError(int errno)
{ if (errno <= MAXERROR) errno = MAXERROR;
  if (ErrDist >= MinErrDist) {
    if (Custom_Error != 0L)
      (*Custom_Error) (errno, S_Line, S_Col, S_Pos);
    errors++;
  }
  ErrDist = 0;
}

static void Get(void)
{ do {
    Sym = S_Get();
    if (Sym <= MAXT) ErrDist ++;
    else {
      /* Empty Stmt */ ;
      S_NextPos  = S_Pos;  S_NextCol = S_Col;
      S_NextLine = S_Line; S_NextLen = S_Len;
    }
  } while (Sym > MAXT);
}

static int In (unsigned short int *SymbolSet, int i)
{ return SymbolSet[i / NSETBITS] & (1 << (i % NSETBITS)); }

static void Expect (int n)
{ if (Sym == n) Get(); else GenError(n); }

static void ExpectWeak (int n, int follow)
{ if (Sym == n) Get();
  else {
    GenError(n);
    while (!(In(SymSet[follow], Sym) || In(SymSet[0], Sym))) Get();
  }
}

static int WeakSeparator (int n, int syFol, int repFol)
{ unsigned short int s[MAXSYM];
  int i;

  if (Sym == n) { Get(); return 1; }
  if (In(SymSet[repFol], Sym)) return 0;
  for (i = 0; i < MAXSYM; i++)
    s[i] = SymSet[0][i] | SymSet[syFol][i] | SymSet[repFol][i];
  GenError(n);
  while (!In(s, Sym)) Get();
  return In(SymSet[syFol], Sym);
}

int Successful(void)
{ return errors == 0; }

/* Productions */

static void XRef(void)
{
	Create();
	Expect(PROGRAMSym);
	Ident();
	Add(Name, S_Line, DEFINED);
	Expect(SemicolonSym);
	Body();
	Ident();
	Add(Name, S_Line, APPLIED);
	Expect(PointSym);
}

static void Ident(void)
{
	Expect(identSym);
	LexName(Name, sizeof(Name)-1);
}

static void Body(void)
{
	while (Sym == VARSym ||
	       Sym == PROCEDURESym) {
		if (Sym == VARSym) {
			Get();
			while (Sym == identSym) {
				Ident();
				Expect(ColonSym);
				Add(Name, S_Line, DEFINED);
				TypeId();
				Expect(SemicolonSym);
			}
		} else if (Sym == PROCEDURESym) {
			Get();
			Ident();
			Add(Name, S_Line, DEFINED);
			Expect(SemicolonSym);
			Body();
			Ident();
			Add(Name, S_Line, APPLIED);
			Expect(SemicolonSym);
		} else GenError(31);
	}
	Expect(BEGINSym);
	StatSeq();
	Expect(ENDSym);
}

static void TypeId(void)
{
	if (Sym == INTEGERSym) {
		Get();
		Add("INTEGER", S_Line, APPLIED);
	} else if (Sym == BOOLEANSym) {
		Get();
		Add("BOOLEAN", S_Line, APPLIED);
	} else GenError(32);
}

static void StatSeq(void)
{
	Stat();
	while (Sym == SemicolonSym) {
		Get();
		Stat();
	}
}

static void Stat(void)
{
	if (Sym == identSym ||
	    Sym == IFSym ||
	    Sym == WHILESym ||
	    Sym >= READSym && Sym <= WRITESym) {
		switch (Sym) {
			case identSym:  
				Ident();
				Add(Name, S_Line, APPLIED);
				if (Sym == ColonEqualSym) {
					Get();
					Expression();
				} else if (Sym == SemicolonSym ||
				           Sym == ENDSym ||
				           Sym == ELSESym) {
					/* Empty */
				} else GenError(33);
				break;
			case IFSym:  
				Get();
				Expression();
				Expect(THENSym);
				StatSeq();
				if (Sym == ELSESym) {
					Get();
					StatSeq();
				}
				Expect(ENDSym);
				break;
			case WHILESym:  
				Get();
				Expression();
				Expect(DOSym);
				StatSeq();
				Expect(ENDSym);
				break;
			case READSym:  
				Get();
				Ident();
				Add(Name, S_Line, APPLIED);
				break;
			case WRITESym:  
				Get();
				Expression();
				break;
			default :GenError(34); break;
		}
	}
}

static void Expression(void)
{
	SimExpr();
	if (Sym >= EqualSym && Sym <= GreaterSym) {
		RelOp();
		SimExpr();
	}
}

static void SimExpr(void)
{
	Term();
	while (Sym == MinusSym ||
	       Sym == PlusSym) {
		AddOp();
		Term();
	}
}

static void RelOp(void)
{
	if (Sym == EqualSym) {
		Get();
	} else if (Sym == LessSym) {
		Get();
	} else if (Sym == GreaterSym) {
		Get();
	} else GenError(35);
}

static void Term(void)
{
	Factor();
	while (Sym >= StarSym && Sym <= SlashSym) {
		MulOp();
		Factor();
	}
}

static void AddOp(void)
{
	if (Sym == PlusSym) {
		Get();
	} else if (Sym == MinusSym) {
		Get();
	} else GenError(36);
}

static void Factor(void)
{
	switch (Sym) {
		case identSym:  
			Ident();
			Add(Name, S_Line, APPLIED);
			break;
		case TRUESym:  
			Get();
			Add("TRUE", S_Line, APPLIED);
			break;
		case FALSESym:  
			Get();
			Add("FALSE", S_Line, APPLIED);
			break;
		case numberSym:  
			Get();
			break;
		case MinusSym:  
			Get();
			Factor();
			break;
		default :GenError(37); break;
	}
}

static void MulOp(void)
{
	if (Sym == StarSym) {
		Get();
	} else if (Sym == SlashSym) {
		Get();
	} else GenError(38);
}



void Parse(void)
{ S_Reset(); Get();
  XRef();
}

