//****************************************************************
//   CPLUS2\PARSER_C.FRM
//   Coco/R C++ Support Frames.
//   Author: Frankie Arzu <farzu@uvg.edu.gt>
//
//   Jun 12, 1996  Version 1.06
//      Many fixes and suggestions thanks to
//      Pat Terry <p.terry@.ru.ac.za>
//   Oct 11, 1997  Version 1.07
//      Eliminated redundant ";" in some places
//****************************************************************

#include "prettyc.hpp"

#include "prettypr.hpp"
#include "prettys.hpp"

/*--------------------------------------------------------------------------*/



#include "prettyp.hpp"
#include "prettys.hpp"

unsigned short int prettyParser::SymSet[][MAXSYM] = {
  /*EOF_Sym */
  {0x1,0x0},
  
  {0x0}
};

void prettyParser::Get()
{ do {
    Sym = Scanner->Get();
    Scanner->NextSym.SetSym(Sym);
    if (Sym <= MAXT) Error->ErrorDist ++;
    else {
      /* Empty Stmt */ ;
      Scanner->NextSym = Scanner->CurrSym;
    }
  } while (Sym > MAXT);
}

void prettyParser::ExpectWeak(int n, int follow)
{ if (Sym == n) Get();
  else {
    GenError(n);
    while (!(In(SymSet[follow], Sym) || In(SymSet[0], Sym))) Get();
  }
}

int prettyParser::WeakSeparator(int n, int syFol, int repFol)
{ unsigned short int s[MAXSYM];
  int i;

  if (Sym == n) { Get(); return 1; }
  if (In(SymSet[repFol], Sym)) return 0;

  for (i = 0; i < MAXSYM; i++)
    s[i] = SymSet[0][i] | SymSet[syFol][i] | SymSet[repFol][i];
  GenError(n);
  while (!In(s, Sym)) Get();
  return In(SymSet[syFol], Sym);
}

void prettyParser::Pretty()
{
	Expect(PROGRAMSym);
	Append("PROGRAM ");
	Ident();
	Expect(SemicolonSym);
	Append(";"); IndentNextLine();
	Body();
	Ident();
	Expect(PointSym);
	Append(".");
}

void prettyParser::Ident()
{
	Expect(identSym);
	char IdentName[20];
	LexName(IdentName, 19);
	Append(IdentName);
}

void prettyParser::Body()
{
	while (Sym == VARSym ||
	       Sym == PROCEDURESym) {
		if (Sym == VARSym) {
			Get();
			Append("VAR"); IndentNextLine();
			while (Sym == identSym) {
				Ident();
				Expect(ColonSym);
				Append(" : ");
				TypeId();
				Expect(SemicolonSym);
				Append(";"); NewLine();
			}
			ExdentNextLine();
		} else if (Sym == PROCEDURESym) {
			Get();
			NewLine(); Append("PROCEDURE ");
			Ident();
			Expect(SemicolonSym);
			Append(";"); IndentNextLine();
			Body();
			Ident();
			Expect(SemicolonSym);
			Append(";"); ExdentNextLine();
			NewLine();
		} else GenError(31);
	}
	Expect(BEGINSym);
	Append("BEGIN");
	StatSeq();
	Expect(ENDSym);
	Append("END ");
}

void prettyParser::TypeId()
{
	if (Sym == INTEGERSym) {
		Get();
		Append("INTEGER");
	} else if (Sym == BOOLEANSym) {
		Get();
		Append("BOOLEAN");
	} else GenError(32);
}

void prettyParser::StatSeq()
{
	IndentNextLine();
	Stat();
	while (Sym == SemicolonSym) {
		Get();
		Append(";"); NewLine();
		Stat();
	}
	ExdentNextLine();
}

void prettyParser::Stat()
{
	if (Sym == identSym ||
	    Sym == IFSym ||
	    Sym == WHILESym ||
	    Sym >= READSym && Sym <= WRITESym) {
		switch (Sym) {
			case identSym:  
				Ident();
				if (Sym == ColonEqualSym) {
					Get();
					Append(" := ");
					Expression();
				} else if (Sym == SemicolonSym ||
				           Sym == ENDSym ||
				           Sym == ELSESym) {
					/* Empty */
				} else GenError(33);
				break;
			case IFSym:  
				Get();
				Append("IF ");
				Expression();
				Expect(THENSym);
				Append(" THEN ");
				StatSeq();
				if (Sym == ELSESym) {
					Get();
					Append("ELSE ");
					StatSeq();
				}
				Expect(ENDSym);
				Append("END");
				break;
			case WHILESym:  
				Get();
				Append("WHILE ");
				Expression();
				Expect(DOSym);
				Append(" DO");
				StatSeq();
				Expect(ENDSym);
				Append("END");
				break;
			case READSym:  
				Get();
				Append("READ ");
				Ident();
				break;
			case WRITESym:  
				Get();
				Append("WRITE ");
				Expression();
				break;
			default :GenError(34); break;
		}
	}
}

void prettyParser::Expression()
{
	SimExpr();
	if (Sym >= EqualSym && Sym <= GreaterSym) {
		RelOp();
		SimExpr();
	}
}

void prettyParser::SimExpr()
{
	Term();
	while (Sym == MinusSym ||
	       Sym == PlusSym) {
		AddOp();
		Term();
	}
}

void prettyParser::RelOp()
{
	if (Sym == EqualSym) {
		Get();
		Append(" = ");
	} else if (Sym == LessSym) {
		Get();
		Append(" < ");
	} else if (Sym == GreaterSym) {
		Get();
		Append(" > ");
	} else GenError(35);
}

void prettyParser::Term()
{
	Factor();
	while (Sym >= StarSym && Sym <= SlashSym) {
		MulOp();
		Factor();
	}
}

void prettyParser::AddOp()
{
	if (Sym == PlusSym) {
		Get();
		Append(" + ");
	} else if (Sym == MinusSym) {
		Get();
		Append(" - ");
	} else GenError(36);
}

void prettyParser::Factor()
{
	switch (Sym) {
		case identSym:  
			Ident();
			break;
		case TRUESym:  
			Get();
			Append("TRUE ");
			break;
		case FALSESym:  
			Get();
			Append("FALSE ");
			break;
		case numberSym:  
			Get();
			char Num[20];
			LexString(Num, 19);
			Append(Num);
			break;
		case MinusSym:  
			Get();
			Append(" - ");
			Factor();
			break;
		default :GenError(37); break;
	}
}

void prettyParser::MulOp()
{
	if (Sym == StarSym) {
		Get();
		Append(" * ");
	} else if (Sym == SlashSym) {
		Get();
		Append(" / ");
	} else GenError(38);
}



void prettyParser::Parse()
{ Scanner->Reset(); Get();
  Pretty();
}

