//****************************************************************
//   CPLUS2\PARSER_C.FRM
//   Coco/R C++ Support Frames.
//   Author: Frankie Arzu <farzu@uvg.edu.gt>
//
//   Jun 12, 1996  Version 1.06
//      Many fixes and suggestions thanks to
//      Pat Terry <p.terry@.ru.ac.za>
//   Oct 11, 1997  Version 1.07
//      Eliminated redundant ";" in some places
//****************************************************************

#include "tastec.hpp"

#include "tl.hpp"
#include "tc.hpp"
#include <string.h>

static void StringToVal (char *s, int &val)
{ int n = 0;
  while (*s) n = 10 * n + (*s++ - '0');
  val = n;
}

extern SymTable *Table;
extern Machine *Emulator;

/*--------------------------------------------------------------------------*/



#include "tastep.hpp"
#include "tastes.hpp"

unsigned short int tasteParser::SymSet[][MAXSYM] = {
  /*EOF_Sym */
  {0x1,0x0},
  
  {0x0}
};

void tasteParser::Get()
{ do {
    Sym = Scanner->Get();
    Scanner->NextSym.SetSym(Sym);
    if (Sym <= MAXT) Error->ErrorDist ++;
    else {
      /* Empty Stmt */ ;
      Scanner->NextSym = Scanner->CurrSym;
    }
  } while (Sym > MAXT);
}

void tasteParser::ExpectWeak(int n, int follow)
{ if (Sym == n) Get();
  else {
    GenError(n);
    while (!(In(SymSet[follow], Sym) || In(SymSet[0], Sym))) Get();
  }
}

int tasteParser::WeakSeparator(int n, int syFol, int repFol)
{ unsigned short int s[MAXSYM];
  int i;

  if (Sym == n) { Get(); return 1; }
  if (In(SymSet[repFol], Sym)) return 0;

  for (i = 0; i < MAXSYM; i++)
    s[i] = SymSet[0][i] | SymSet[syFol][i] | SymSet[repFol][i];
  GenError(n);
  while (!In(s, Sym)) Get();
  return In(SymSet[syFol], Sym);
}

void tasteParser::Taste()
{
	Name name, progName;
	Object obj;
	Expect(PROGRAMSym);
	Ident(progName);
	Expect(SemicolonSym);
	Emulator->progStart = Emulator->pc;
	Body();
	Ident(name);
	if (strcmp(name, progName) != 0)
	  SemError(119);
	Emulator->Emit(HALTc);
	Expect(PointSym);
}

void tasteParser::Ident(Name name)
{
	Expect(identSym);
	LexName(name, sizeof(name)-1);
}

void tasteParser::Body()
{
	int fix, type;
	Name name, name1;
	Object obj;
	Table->EnterScope();
	fix = Emulator->pc + 1; Emulator->Emit2(JMP, 0);
	while (Sym == VARSym ||
	       Sym == PROCEDURESym) {
		if (Sym == VARSym) {
			Get();
			while (Sym == identSym) {
				Ident(name);
				Expect(ColonSym);
				obj = Table->NewObj(name, VARS);
				TypeId((*obj).type);
				Expect(SemicolonSym);
			}
		} else if (Sym == PROCEDURESym) {
			Get();
			Ident(name);
			Expect(SemicolonSym);
			obj = Table->NewObj(name, PROCS );
			obj->adr = Emulator->pc;
			Body();
			Ident(name1);
			Emulator->Emit(RET);
			if (strcmp(name, name1) != 0)
			  SemError(119);
			Expect(SemicolonSym);
		} else GenError(30);
	}
	Expect(BEGINSym);
	Emulator->Fixup(fix); Emulator->Emit2(RES, Table->DataSpace());
	StatSeq();
	Expect(ENDSym);
	Table->LeaveScope();
}

void tasteParser::TypeId(int &type)
{
	type = UNDEF;
	if (Sym == INTEGERSym) {
		Get();
		type = INT;
	} else if (Sym == BOOLEANSym) {
		Get();
		type = BOOL;
	} else GenError(31);
}

void tasteParser::StatSeq()
{
	Stat();
	while (Sym == SemicolonSym) {
		Get();
		Stat();
	}
}

void tasteParser::Stat()
{
	int  type;
	Name name;
	Object obj;
	int fix, fix2, loopstart;
	if (Sym == identSym ||
	    Sym == IFSym ||
	    Sym == WHILESym ||
	    Sym >= READSym && Sym <= WRITESym) {
		switch (Sym) {
			case identSym:  
				Ident(name);
				obj = Table->Obj(name);
				if (Sym == ColonSym) {
					Get();
					Expect(EqualSym);
					if (obj->kind != VARS) SemError(123);
					Expression(type);
					if (type != obj->type) SemError(121);
					Emulator->Emit3(STO, Table->curLevel-obj->level, obj->adr);
				} else if (Sym == SemicolonSym ||
				           Sym == ENDSym ||
				           Sym == ELSESym) {
					if (obj->kind != PROCS) SemError(124);
					Emulator->Emit3(CALL, Table->curLevel-obj->level, obj->adr);
				} else GenError(32);
				break;
			case IFSym:  
				Get();
				Expression(type);
				if (type != BOOL) SemError(122);
				fix = Emulator->pc + 1; Emulator->Emit2(FJMP, 0);
				Expect(THENSym);
				StatSeq();
				if (Sym == ELSESym) {
					Get();
					fix2 = Emulator->pc + 1; Emulator->Emit2(JMP, 0);
					Emulator->Fixup(fix); fix = fix2;
					StatSeq();
				}
				Expect(ENDSym);
				Emulator->Fixup(fix);
				break;
			case WHILESym:  
				Get();
				loopstart = Emulator->pc;
				Expression(type);
				if (type != BOOL) SemError(122);
				fix = Emulator->pc + 1; Emulator->Emit2(FJMP, 0);
				Expect(DOSym);
				StatSeq();
				Emulator->Emit2(JMP, loopstart); Emulator->Fixup(fix);
				Expect(ENDSym);
				break;
			case READSym:  
				Get();
				Ident(name);
				obj = Table->Obj(name);
				if (obj->type != INT) SemError(120);
				Emulator->Emit3(READ, Table->curLevel-obj->level, obj->adr);
				break;
			case WRITESym:  
				Get();
				Expression(type);
				if (type != INT) SemError(120);
				Emulator->Emit(WRITE);
				break;
			default :GenError(33); break;
		}
	}
}

void tasteParser::Expression(int &type)
{
	int type1, op;
	SimExpr(type);
	if (Sym == EqualSym ||
	    Sym >= LessSym && Sym <= GreaterSym) {
		RelOp(op);
		SimExpr(type1);
		if (type != type1) SemError(121);
		Emulator->Emit(op); type = BOOL;
	}
}

void tasteParser::SimExpr(int &type)
{
	int type1, op;
	Term(type);
	while (Sym == MinusSym ||
	       Sym == PlusSym) {
		AddOp(op);
		Term(type1);
		if (type != INT || type1 != INT) SemError(120);
		Emulator->Emit(op);
	}
}

void tasteParser::RelOp(int &op)
{
	if (Sym == EqualSym) {
		op = -1;
		Get();
		op = EQU;
	} else if (Sym == LessSym) {
		Get();
		op = LSS;
	} else if (Sym == GreaterSym) {
		Get();
		op = GTR;
	} else GenError(34);
}

void tasteParser::Term(int &type)
{
	int type1, op;
	Factor(type);
	while (Sym >= StarSym && Sym <= SlashSym) {
		MulOp(op);
		Factor(type1);
		if (type != INT || type1 != INT) SemError(120);
		Emulator->Emit(op);
	}
}

void tasteParser::AddOp(int &op)
{
	op = -1;
	if (Sym == PlusSym) {
		Get();
		op = PLUS;
	} else if (Sym == MinusSym) {
		Get();
		op = MINUS;
	} else GenError(35);
}

void tasteParser::Factor(int &type)
{
	int val, n;
	Object obj;
	Name name;
	switch (Sym) {
		case identSym:  
			Ident(name);
			obj = Table->Obj(name); type = obj->type;
			if (obj->kind == VARS)
			  Emulator->Emit3(LOAD, Table->curLevel-obj->level, obj->adr);
			else SemError(123);
			break;
		case TRUESym:  
			Get();
			Emulator->Emit2(LIT, 1); type = BOOL;
			break;
		case FALSESym:  
			Get();
			Emulator->Emit2(LIT, 0); type = BOOL;
			break;
		case numberSym:  
			Get();
			LexString(name, sizeof(name)-1);
			StringToVal(name, n);
			Emulator->Emit2(LIT, n); type = INT;
			break;
		case MinusSym:  
			Get();
			Factor(type);
			if (type != INT) { SemError(120); type = INT; }
			Emulator->Emit(NEG);
			break;
		default :GenError(36); break;
	}
}

void tasteParser::MulOp(int &op)
{
	op = -1;
	if (Sym == StarSym) {
		Get();
		op = TIMES;
	} else if (Sym == SlashSym) {
		Get();
		op = SLASH;
	} else GenError(37);
}



void tasteParser::Parse()
{ Scanner->Reset(); Get();
  Taste();
}

