//****************************************************************
//   CPLUS2\SCAN_C.FRM
//   Coco/R C++ Support Frames.
//   Author: Frankie Arzu <farzu@uvg.edu.gt>
//
//   Jun 12, 1996  Version 1.06
//      Many fixes and suggestions thanks to
//      Pat Terry <p.terry@.ru.ac.za>
//   Oct 31, 1999  Version 1.14
//      LeftContext Support
//   Mar 24, 2000  Version 1.15
//      LeftContext Support no longer needed
//****************************************************************

#include "tastec.hpp"
#include "tastes.hpp"

#define Scan_Ch        Ch
#define Scan_NextCh    NextCh
#define Scan_ComEols   ComEols
#define Scan_CurrLine  CurrLine
#define Scan_CurrCol   CurrCol
#define Scan_LineStart LineStart
#define Scan_BuffPos   BuffPos
#define Scan_NextLen   NextSym.Len

int tasteScanner::STATE0[] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
                  0,0,0,0,0,0,0,0,0,0,0,0,8,10,0,7,4,9,2,2,2,2,2,2,2,2,2,2,5,3,
                  11,6,12,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
                  1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
                  1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
                  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
                  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
                  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
                  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

int tasteScanner::CheckLiteral(int id)
{ char c;
  c =  CurrentCh(NextSym.Pos);
  if (IgnoreCase) c = Upcase(c);
  switch (c) {
  	case 'B':
  		if (EqualStr("BEGIN")) return BEGINSym;
  		if (EqualStr("BOOLEAN")) return BOOLEANSym;
  		break;
  	case 'D':
  		if (EqualStr("DO")) return DOSym;
  		break;
  	case 'E':
  		if (EqualStr("END")) return ENDSym;
  		if (EqualStr("ELSE")) return ELSESym;
  		break;
  	case 'F':
  		if (EqualStr("FALSE")) return FALSESym;
  		break;
  	case 'I':
  		if (EqualStr("INTEGER")) return INTEGERSym;
  		if (EqualStr("IF")) return IFSym;
  		break;
  	case 'P':
  		if (EqualStr("PROGRAM")) return PROGRAMSym;
  		if (EqualStr("PROCEDURE")) return PROCEDURESym;
  		break;
  	case 'R':
  		if (EqualStr("READ")) return READSym;
  		break;
  	case 'T':
  		if (EqualStr("THEN")) return THENSym;
  		if (EqualStr("TRUE")) return TRUESym;
  		break;
  	case 'V':
  		if (EqualStr("VAR")) return VARSym;
  		break;
  	case 'W':
  		if (EqualStr("WHILE")) return WHILESym;
  		if (EqualStr("WRITE")) return WRITESym;
  		break;
  
  }
  return id;
}

int tasteScanner::Comment()
{ int Level, StartLine, OldCol;
  long OldLineStart;

  Level = 1; StartLine = CurrLine;
  OldLineStart = LineStart; OldCol = CurrCol;
  if (Scan_Ch == '(') { /* 1 */
  	Scan_NextCh();
  	if (Scan_Ch == '*') { /* 2 */
  		Scan_NextCh();
  		while (1) {
  			if (Scan_Ch== '*') { /* 5 */
  				Scan_NextCh();
  				if (Scan_Ch == ')') { /* 6 */
  					Level--; Scan_NextCh(); Scan_ComEols = Scan_CurrLine - StartLine;
  					if(Level == 0) return 1;
  				} /* 6 */ 
  			} else  /* 5 */
  			if (Scan_Ch == '(') {
  				Scan_NextCh();
  				if (Scan_Ch == '*') { Level++; Scan_NextCh(); }
  			} else /* 5 */
  			if (Scan_Ch == EOF_CHAR) return 0;
  			else Scan_NextCh();
  		} /* while */
  	} else { /* 2 */
  		if (Scan_Ch == LF_CHAR) { Scan_CurrLine--; Scan_LineStart = OldLineStart; }
  		Scan_BuffPos -= 2; Scan_CurrCol = OldCol - 1; Scan_NextCh();
  	} /* 2 */
  } /* 1*/
  
  return 0;
}

int tasteScanner::Get()
{ int state, ctx;

  start:
    while (Scan_Ch >= 9 && Scan_Ch <= 10 ||
           Scan_Ch == 13 ||
           Scan_Ch == ' ') Scan_NextCh();
    if ((Scan_Ch == '(') && Comment()) goto start;

    CurrSym = NextSym;
    NextSym.Init(0, CurrLine, CurrCol - 1, BuffPos, 0);
    NextSym.Len  = 0; ctx = 0;

    if (Ch == EOF_CHAR) return EOF_Sym;
    state = STATE0[Ch];
    while(1) {
      Scan_NextCh(); NextSym.Len++;
      switch (state) {
       /* State 0; valid STATE0 Table
      case 0:
      	if (Scan_Ch >= 'A' && Scan_Ch <= 'Z' ||
          Scan_Ch >= 'a' && Scan_Ch <= 'z') state = 1; else
      	if (Scan_Ch >= '0' && Scan_Ch <= '9') state = 2; else
      	if (Scan_Ch == ';') state = 3; else
      	if (Scan_Ch == '.') state = 4; else
      	if (Scan_Ch == ':') state = 5; else
      	if (Scan_Ch == '=') state = 6; else
      	if (Scan_Ch == '-') state = 7; else
      	if (Scan_Ch == '*') state = 8; else
      	if (Scan_Ch == '/') state = 9; else
      	if (Scan_Ch == '+') state = 10; else
      	if (Scan_Ch == '<') state = 11; else
      	if (Scan_Ch == '>') state = 12; else
      	return No_Sym;
      	break;
       --------- End State0 --------- */
      case 1:
      	if (Scan_Ch >= '0' && Scan_Ch <= '9' ||
      	    Scan_Ch >= 'A' && Scan_Ch <= 'Z' ||
      	    Scan_Ch >= 'a' && Scan_Ch <= 'z') /*same state*/; else
      	return CheckLiteral(identSym);
      	break;
      case 2:
      	if (Scan_Ch >= '0' && Scan_Ch <= '9') /*same state*/; else
      	return numberSym;
      	break;
      case 3:
      	return SemicolonSym;
      case 4:
      	return PointSym;
      case 5:
      	return ColonSym;
      case 6:
      	return EqualSym;
      case 7:
      	return MinusSym;
      case 8:
      	return StarSym;
      case 9:
      	return SlashSym;
      case 10:
      	return PlusSym;
      case 11:
      	return LessSym;
      case 12:
      	return GreaterSym;
      
      default: return No_Sym; /* Scan_NextCh already done */
      }
    }
}

