//****************************************************************
//   CPLUS2\PARSER_C.FRM
//   Coco/R C++ Support Frames.
//   Author: Frankie Arzu <farzu@uvg.edu.gt>
//
//   Jun 12, 1996  Version 1.06
//      Many fixes and suggestions thanks to
//      Pat Terry <p.terry@.ru.ac.za>
//   Oct 11, 1997  Version 1.07
//      Eliminated redundant ";" in some places
//****************************************************************

#include "xrefc.hpp"

#include "crossref.hpp"
#define  APPLIED  0
#define  DEFINED  1

static  char Name[15];
RefTable Table;

/*------------------------------------------------------------------------*/



#include "xrefp.hpp"
#include "xrefs.hpp"

unsigned short int xrefParser::SymSet[][MAXSYM] = {
  /*EOF_Sym */
  {0x1,0x0},
  
  {0x0}
};

void xrefParser::Get()
{ do {
    Sym = Scanner->Get();
    Scanner->NextSym.SetSym(Sym);
    if (Sym <= MAXT) Error->ErrorDist ++;
    else {
      /* Empty Stmt */ ;
      Scanner->NextSym = Scanner->CurrSym;
    }
  } while (Sym > MAXT);
}

void xrefParser::ExpectWeak(int n, int follow)
{ if (Sym == n) Get();
  else {
    GenError(n);
    while (!(In(SymSet[follow], Sym) || In(SymSet[0], Sym))) Get();
  }
}

int xrefParser::WeakSeparator(int n, int syFol, int repFol)
{ unsigned short int s[MAXSYM];
  int i;

  if (Sym == n) { Get(); return 1; }
  if (In(SymSet[repFol], Sym)) return 0;

  for (i = 0; i < MAXSYM; i++)
    s[i] = SymSet[0][i] | SymSet[syFol][i] | SymSet[repFol][i];
  GenError(n);
  while (!In(s, Sym)) Get();
  return In(SymSet[syFol], Sym);
}

void xrefParser::XRef()
{
	Expect(PROGRAMSym);
	Ident();
	Table.Add(Name, Scanner->CurrSym.Line, DEFINED);
	Expect(SemicolonSym);
	Body();
	Ident();
	Table.Add(Name, Scanner->CurrSym.Line, APPLIED);
	Expect(PointSym);
}

void xrefParser::Ident()
{
	Expect(identSym);
	LexName(Name, sizeof(Name)-1);
}

void xrefParser::Body()
{
	while (Sym == VARSym ||
	       Sym == PROCEDURESym) {
		if (Sym == VARSym) {
			Get();
			while (Sym == identSym) {
				Ident();
				Expect(ColonSym);
				Table.Add(Name, Scanner->CurrSym.Line, DEFINED);
				TypeId();
				Expect(SemicolonSym);
			}
		} else if (Sym == PROCEDURESym) {
			Get();
			Ident();
			Table.Add(Name, Scanner->CurrSym.Line, DEFINED);
			Expect(SemicolonSym);
			Body();
			Ident();
			Table.Add(Name, Scanner->CurrSym.Line, APPLIED);
			Expect(SemicolonSym);
		} else GenError(31);
	}
	Expect(BEGINSym);
	StatSeq();
	Expect(ENDSym);
}

void xrefParser::TypeId()
{
	if (Sym == INTEGERSym) {
		Get();
		Table.Add("INTEGER", Scanner->CurrSym.Line, APPLIED);
	} else if (Sym == BOOLEANSym) {
		Get();
		Table.Add("BOOLEAN", Scanner->CurrSym.Line, APPLIED);
	} else GenError(32);
}

void xrefParser::StatSeq()
{
	Stat();
	while (Sym == SemicolonSym) {
		Get();
		Stat();
	}
}

void xrefParser::Stat()
{
	if (Sym == identSym ||
	    Sym == IFSym ||
	    Sym == WHILESym ||
	    Sym >= READSym && Sym <= WRITESym) {
		switch (Sym) {
			case identSym:  
				Ident();
				Table.Add(Name, Scanner->CurrSym.Line, APPLIED);
				if (Sym == ColonEqualSym) {
					Get();
					Expression();
				} else if (Sym == SemicolonSym ||
				           Sym == ENDSym ||
				           Sym == ELSESym) {
					/* Empty */
				} else GenError(33);
				break;
			case IFSym:  
				Get();
				Expression();
				Expect(THENSym);
				StatSeq();
				if (Sym == ELSESym) {
					Get();
					StatSeq();
				}
				Expect(ENDSym);
				break;
			case WHILESym:  
				Get();
				Expression();
				Expect(DOSym);
				StatSeq();
				Expect(ENDSym);
				break;
			case READSym:  
				Get();
				Ident();
				Table.Add(Name, Scanner->CurrSym.Line, APPLIED);
				break;
			case WRITESym:  
				Get();
				Expression();
				break;
			default :GenError(34); break;
		}
	}
}

void xrefParser::Expression()
{
	SimExpr();
	if (Sym >= EqualSym && Sym <= GreaterSym) {
		RelOp();
		SimExpr();
	}
}

void xrefParser::SimExpr()
{
	Term();
	while (Sym == MinusSym ||
	       Sym == PlusSym) {
		AddOp();
		Term();
	}
}

void xrefParser::RelOp()
{
	if (Sym == EqualSym) {
		Get();
	} else if (Sym == LessSym) {
		Get();
	} else if (Sym == GreaterSym) {
		Get();
	} else GenError(35);
}

void xrefParser::Term()
{
	Factor();
	while (Sym >= StarSym && Sym <= SlashSym) {
		MulOp();
		Factor();
	}
}

void xrefParser::AddOp()
{
	if (Sym == PlusSym) {
		Get();
	} else if (Sym == MinusSym) {
		Get();
	} else GenError(36);
}

void xrefParser::Factor()
{
	switch (Sym) {
		case identSym:  
			Ident();
			Table.Add(Name, Scanner->CurrSym.Line, APPLIED);
			break;
		case TRUESym:  
			Get();
			Table.Add("TRUE", Scanner->CurrSym.Line, APPLIED);
			break;
		case FALSESym:  
			Get();
			Table.Add("FALSE", Scanner->CurrSym.Line, APPLIED);
			break;
		case numberSym:  
			Get();
			break;
		case MinusSym:  
			Get();
			Factor();
			break;
		default :GenError(37); break;
	}
}

void xrefParser::MulOp()
{
	if (Sym == StarSym) {
		Get();
	} else if (Sym == SlashSym) {
		Get();
	} else GenError(38);
}



void xrefParser::Parse()
{ Scanner->Reset(); Get();
  XRef();
}

